@@ -1,48 +1,41 @@
 use crate::common::*;
 
-#[derive(Debug, PartialEq)]
+use serde::Serialize;
+
+#[derive(Debug, PartialEq, Serialize)]
 pub(crate) struct Justfile<'src> {
-  pub(crate) recipes:     Table<'src, Rc<Recipe<'src>>>,
+  pub(crate) aliases: Table<'src, Alias<'src>>,
   pub(crate) assignments: Table<'src, Assignment<'src>>,
-  pub(crate) aliases:     Table<'src, Alias<'src>>,
-  pub(crate) settings:    Settings<'src>,
-  pub(crate) warnings:    Vec<Warning>,
+  #[serde(serialize_with = "keyed::serialize_option")]
+  pub(crate) first: Option<Rc<Recipe<'src>>>,
+  pub(crate) recipes: Table<'src, Rc<Recipe<'src>>>,
+  pub(crate) settings: Settings<'src>,
+  pub(crate) warnings: Vec<Warning>,
 }
 
 impl<'src> Justfile<'src> {
-  pub(crate) fn first(&self) -> Option<&Recipe> {
-    let mut first: Option<&Recipe<Dependency>> = None;
-    for recipe in self.recipes.values() {
-      if let Some(first_recipe) = first {
-        if recipe.line_number() < first_recipe.line_number() {
-          first = Some(recipe);
-        }
-      } else {
-        first = Some(recipe);
-      }
-    }
-    first
-  }
-
   pub(crate) fn count(&self) -> usize {
     self.recipes.len()
   }
 
-  pub(crate) fn suggest_recipe(&self, input: &str) -> Option<Suggestion> {
+  pub(crate) fn suggest_recipe(&self, input: &str) -> Option<Suggestion<'src>> {
     let mut suggestions = self
       .recipes
       .keys()
       .map(|name| {
-        (edit_distance(name, input), Suggestion {
-          name,
-          target: None,
-        })
+        (
+          edit_distance(name, input),
+          Suggestion { name, target: None },
+        )
       })
       .chain(self.aliases.iter().map(|(name, alias)| {
-        (edit_distance(name, input), Suggestion {
-          name,
-          target: Some(alias.target.name.lexeme()),
-        })
+        (
+          edit_distance(name, input),
+          Suggestion {
+            name,
+            target: Some(alias.target.name.lexeme()),
+          },
+        )
       }))
       .filter(|(distance, _suggestion)| distance < &3)
       .collect::<Vec<(usize, Suggestion)>>();
@@ -54,15 +47,15 @@ impl<'src> Justfile<'src> {
       .next()
   }
 
-  pub(crate) fn suggest_variable(&self, input: &str) -> Option<Suggestion> {
+  pub(crate) fn suggest_variable(&self, input: &str) -> Option<Suggestion<'src>> {
     let mut suggestions = self
       .assignments
       .keys()
       .map(|name| {
-        (edit_distance(name, input), Suggestion {
-          name,
-          target: None,
-        })
+        (
+          edit_distance(name, input),
+          Suggestion { name, target: None },
+        )
       })
       .filter(|(distance, _suggestion)| distance < &3)
       .collect::<Vec<(usize, Suggestion)>>();
@@ -74,21 +67,25 @@ impl<'src> Justfile<'src> {
       .next()
   }
 
-  pub(crate) fn run<'run>(
-    &'run self,
-    config: &'run Config,
-    search: &'run Search,
-    overrides: &'run BTreeMap<String, String>,
-    arguments: &'run [String],
-  ) -> RunResult<'run, ()> {
+  pub(crate) fn run(
+    &self,
+    config: &Config,
+    search: &Search,
+    overrides: &BTreeMap<String, String>,
+    arguments: &[String],
+  ) -> RunResult<'src, ()> {
+    if let Err(error) = InterruptHandler::install(config.verbosity) {
+      warn!("Failed to set CTRL-C handler: {}", error);
+    }
+
     let unknown_overrides = overrides
       .keys()
       .filter(|name| !self.assignments.contains_key(name.as_str()))
-      .map(String::as_str)
-      .collect::<Vec<&str>>();
+      .cloned()
+      .collect::<Vec<String>>();
 
     if !unknown_overrides.is_empty() {
-      return Err(RuntimeError::UnknownOverrides {
+      return Err(Error::UnknownOverrides {
         overrides: unknown_overrides,
       });
     }
@@ -107,12 +104,12 @@ impl<'src> Justfile<'src> {
         if let Some(assignment) = self.assignments.get(name) {
           scope.bind(assignment.export, assignment.name, value.clone());
         } else {
-          unknown_overrides.push(name.as_ref());
+          unknown_overrides.push(name.clone());
         }
       }
 
       if !unknown_overrides.is_empty() {
-        return Err(RuntimeError::UnknownOverrides {
+        return Err(Error::UnknownOverrides {
           overrides: unknown_overrides,
         });
       }
@@ -148,7 +145,7 @@ impl<'src> Justfile<'src> {
         command.export(&self.settings, &dotenv, &scope);
 
         let status = InterruptHandler::guard(|| command.status()).map_err(|io_error| {
-          RuntimeError::CommandInvocation {
+          Error::CommandInvoke {
             binary: binary.clone(),
             arguments: arguments.clone(),
             io_error,
@@ -156,19 +153,23 @@ impl<'src> Justfile<'src> {
         })?;
 
         if !status.success() {
-          process::exit(status.code().unwrap_or(EXIT_FAILURE));
+          return Err(Error::CommandStatus {
+            binary: binary.clone(),
+            arguments: arguments.clone(),
+            status,
+          });
         };
 
         return Ok(());
-      },
+      }
       Subcommand::Evaluate { variable, .. } => {
         if let Some(variable) = variable {
           if let Some(value) = scope.value(variable) {
             print!("{}", value);
           } else {
-            return Err(RuntimeError::EvalUnknownVariable {
+            return Err(Error::EvalUnknownVariable {
               suggestion: self.suggest_variable(&variable),
-              variable:   variable.clone(),
+              variable: variable.clone(),
             });
           }
         } else {
@@ -189,23 +190,23 @@ impl<'src> Justfile<'src> {
         }
 
         return Ok(());
-      },
-      _ => {},
+      }
+      _ => {}
     }
 
     let argvec: Vec<&str> = if !arguments.is_empty() {
       arguments.iter().map(String::as_str).collect()
-    } else if let Some(recipe) = self.first() {
+    } else if let Some(recipe) = &self.first {
       let min_arguments = recipe.min_arguments();
       if min_arguments > 0 {
-        return Err(RuntimeError::DefaultRecipeRequiresArguments {
+        return Err(Error::DefaultRecipeRequiresArguments {
           recipe: recipe.name.lexeme(),
           min_arguments,
         });
       }
       vec![recipe.name()]
     } else {
-      return Err(RuntimeError::NoRecipes);
+      return Err(Error::NoRecipes);
     };
 
     let arguments = argvec.as_slice();
@@ -222,19 +223,19 @@ impl<'src> Justfile<'src> {
           let argument_range = recipe.argument_range();
           let argument_count = cmp::min(tail.len(), recipe.max_arguments());
           if !argument_range.range_contains(&argument_count) {
-            return Err(RuntimeError::ArgumentCountMismatch {
-              recipe:     recipe.name(),
-              parameters: recipe.parameters.iter().collect(),
-              found:      tail.len(),
-              min:        recipe.min_arguments(),
-              max:        recipe.max_arguments(),
+            return Err(Error::ArgumentCountMismatch {
+              recipe: recipe.name(),
+              parameters: recipe.parameters.clone(),
+              found: tail.len(),
+              min: recipe.min_arguments(),
+              max: recipe.max_arguments(),
             });
           }
           grouped.push((recipe, &tail[0..argument_count]));
           tail = &tail[argument_count..];
         }
       } else {
-        missing.push(*argument);
+        missing.push((*argument).to_owned());
       }
       rest = tail;
     }
@@ -245,7 +246,7 @@ impl<'src> Justfile<'src> {
       } else {
         None
       };
-      return Err(RuntimeError::UnknownRecipes {
+      return Err(Error::UnknownRecipes {
         recipes: missing,
         suggestion,
       });
@@ -266,7 +267,7 @@ impl<'src> Justfile<'src> {
     Ok(())
   }
 
-  pub(crate) fn get_alias(&self, name: &str) -> Option<&Alias> {
+  pub(crate) fn get_alias(&self, name: &str) -> Option<&Alias<'src>> {
     self.aliases.get(name)
   }
 
@@ -278,13 +279,13 @@ impl<'src> Justfile<'src> {
       .or_else(|| self.aliases.get(name).map(|alias| alias.target.as_ref()))
   }
 
-  fn run_recipe<'run>(
+  fn run_recipe(
     &self,
-    context: &'run RecipeContext<'src, 'run>,
+    context: &RecipeContext<'src, '_>,
     recipe: &Recipe<'src>,
-    arguments: &[&'run str],
+    arguments: &[&str],
     dotenv: &BTreeMap<String, String>,
-    search: &'run Search,
+    search: &Search,
     ran: &mut BTreeSet<Vec<String>>,
   ) -> RunResult<'src, ()> {
     let (outer, positional) = Evaluator::evaluate_parameters(
@@ -302,7 +303,7 @@ impl<'src> Justfile<'src> {
     let mut evaluator =
       Evaluator::recipe_evaluator(context.config, dotenv, &scope, context.settings, search);
 
-    for Dependency { recipe, arguments } in &recipe.dependencies {
+    for Dependency { recipe, arguments } in recipe.dependencies.iter().take(recipe.priors) {
       let mut invocation = vec![recipe.name().to_owned()];
 
       for argument in arguments {
@@ -321,6 +322,27 @@ impl<'src> Justfile<'src> {
 
     recipe.run(context, dotenv, scope.child(), search, &positional)?;
 
+    {
+      let mut ran = BTreeSet::new();
+
+      for Dependency { recipe, arguments } in recipe.dependencies.iter().skip(recipe.priors) {
+        let mut evaluated = Vec::new();
+
+        for argument in arguments {
+          evaluated.push(evaluator.evaluate_expression(argument)?);
+        }
+
+        self.run_recipe(
+          context,
+          recipe,
+          &evaluated.iter().map(String::as_ref).collect::<Vec<&str>>(),
+          dotenv,
+          search,
+          &mut ran,
+        )?;
+      }
+    }
+
     let mut invocation = vec![recipe.name().to_owned()];
     for argument in arguments.iter().cloned() {
       invocation.push(argument.to_owned());
@@ -330,7 +352,7 @@ impl<'src> Justfile<'src> {
     Ok(())
   }
 
-  pub(crate) fn public_recipes(&self, source_order: bool) -> Vec<&Recipe<Dependency>> {
+  pub(crate) fn public_recipes(&self, source_order: bool) -> Vec<&Recipe<'src, Dependency>> {
     let mut recipes = self
       .recipes
       .values()
@@ -346,8 +368,8 @@ impl<'src> Justfile<'src> {
   }
 }
 
-impl<'src> Display for Justfile<'src> {
-  fn fmt(&self, f: &mut Formatter) -> Result<(), fmt::Error> {
+impl<'src> ColorDisplay for Justfile<'src> {
+  fn fmt(&self, f: &mut Formatter, color: Color) -> Result<(), fmt::Error> {
     let mut items = self.recipes.len() + self.assignments.len() + self.aliases.len();
     for (name, assignment) in &self.assignments {
       if assignment.export {
@@ -367,7 +389,7 @@ impl<'src> Display for Justfile<'src> {
       }
     }
     for recipe in self.recipes.values() {
-      write!(f, "{}", recipe)?;
+      write!(f, "{}", recipe.color_display(color))?;
       items -= 1;
       if items != 0 {
         write!(f, "\n\n")?;
@@ -382,7 +404,7 @@ mod tests {
   use super::*;
 
   use testing::compile;
-  use RuntimeError::*;
+  use Error::*;
 
   run_error! {
     name: unknown_recipes,
@@ -654,11 +676,11 @@ mod tests {
 
   fn test(input: &str, expected: &str) {
     let justfile = compile(input);
-    let actual = format!("{:#}", justfile);
+    let actual = format!("{}", justfile.color_display(Color::never()));
     assert_eq!(actual, expected);
     println!("Re-parsing...");
     let reparsed = compile(&actual);
-    let redumped = format!("{:#}", reparsed);
+    let redumped = format!("{}", reparsed.color_display(Color::never()));
     assert_eq!(redumped, actual);
   }
 
@@ -802,7 +824,7 @@ goodbye := \"y\"
 hello a b c: x y z
     #! blah
     #blarg
-    {{foo + bar}}abc{{goodbye + \"x\"}}xyz
+    {{ foo + bar }}abc{{ goodbye + \"x\" }}xyz
     1
     2
     3
@@ -828,7 +850,7 @@ install:
 
 install:
     #!/bin/sh
-    if [[ -f {{practicum}} ]]; then
+    if [[ -f {{ practicum }} ]]; then
     \treturn
     fi",
   }
@@ -869,7 +891,7 @@ c := a + b + a + b",
     r#"a:
   echo {{  `echo hello` + "blarg"   }} {{   `echo bob`   }}"#,
     r#"a:
-    echo {{`echo hello` + "blarg"}} {{`echo bob`}}"#,
+    echo {{ `echo hello` + "blarg" }} {{ `echo bob` }}"#,
   }
 
   test! {
@@ -895,7 +917,7 @@ c := a + b + a + b",
     "a b c:
   {{b}} {{c}}",
     "a b c:
-    {{b}} {{c}}",
+    {{ b }} {{ c }}",
   }
 
   test! {
@@ -908,7 +930,7 @@ a:
     "x := arch()
 
 a:
-    {{os()}} {{os_family()}}",
+    {{ os() }} {{ os_family() }}",
   }
 
   test! {
@@ -921,7 +943,7 @@ a:
     r#"x := env_var('foo')
 
 a:
-    {{env_var_or_default('foo' + 'bar', 'baz')}} {{env_var(env_var("baz"))}}"#,
+    {{ env_var_or_default('foo' + 'bar', 'baz') }} {{ env_var(env_var("baz")) }}"#,
   }
 
   test! {
